import sys
from math import sqrt, exp
from scipy.constants import pi, N_A, k, h, physical_constants
from importlib.metadata import version

k_eV = physical_constants["Boltzmann constant in eV/K"][0]
atomic_mass = physical_constants["atomic mass constant"][0]


def write_header(path):
    """Write the header for the new Zacros input file.

    Parameters
    ----------
    path: str
        Path of the input file
    """
    with open(path, 'w') as infile:
        infile.write('############################################################################\n')
        infile.write(f"#  Zacros input file generated by ZacrosTools (version {version('zacrostools')+')':<18}  #\n")
        infile.write('#  https://github.com/hprats/ZacrosTools                                   #\n')
        infile.write('#                                                                          #\n')
        infile.write('#  Hector Prats, PhD                                                       #\n')
        infile.write('############################################################################\n\n')


def get_q_vib(temperature, vib_energies):
    """Calculates the vibrational partition function.

    Parameters
    ----------
    temperature: float
        Temperature (in K)
    vib_energies: list of float
        Vibrational energies of the molecule or adsorbate (in meV)

    Returns
    -------
    q_vib: float
        Vibrational partition function

    """
    q_vib = 1.0
    for v in vib_energies:
        q_vib = q_vib * exp(- v / (1000 * 2 * k_eV * temperature)) / (1 - exp(- v / (1000 * k_eV * temperature)))
    return q_vib


def get_q_rot(temperature, inertia_moments: list, sym_number):
    """Calculates the rotational partition function.

    Parameters
    ----------
    temperature: float
        Temperature (in K)
    inertia_moments: list
        Moments of inertia in amu·Å^2. 1 element for linear molecules, 3 elements for non-linear molecules. Can be
        obtained from ase.Atoms.get_moments_of_inertia()
    sym_number: int
        Molecule symmetry number

    Returns
    -------
    q_rot_gas: float
        Rotational partition function

    """

    if len(inertia_moments) == 1:  # linear
        i = inertia_moments[0] * atomic_mass / 1.0e20  # from amu*Å2 to kg*m2
        q_rot_gas = 8 * pi ** 2 * i * k * temperature / (sym_number * h ** 2)
    elif len(inertia_moments) == 3:  # non-linear
        i_a = inertia_moments[0] * atomic_mass / 1.0e20
        i_b = inertia_moments[1] * atomic_mass / 1.0e20
        i_c = inertia_moments[2] * atomic_mass / 1.0e20
        q_rot_gas = (sqrt(pi * i_a * i_b * i_c) / sym_number) * (8 * pi ** 2 * k * temperature / h ** 2) ** (3 / 2)
    else:
        sys.exit(f"Invalid inertia_moments")
    return q_rot_gas


def calc_ads(area_site, molec_mass, temperature, vib_energies_is, vib_energies_ts, vib_energies_fs,
             inertia_moments: list, sym_number, degeneracy=1):
    """Calculates the forward and reverse pre-exponential factors for a reversible activated adsorption.

    Parameters
    ----------
    area_site: float
        Area of an adsorption site (in Å^2)
    molec_mass: float
        Molecular mass (in g/mol) of the gas species
    temperature: float
        Temperature (in K)
    vib_energies_is: list
        Vibrational energies for the initial state (in meV)
    vib_energies_ts: list
        Vibrational energies for the transition state (in meV)
    vib_energies_fs: list
        Vibrational energies for the final state (in meV)
    inertia_moments: list
        Moments of inertia for the gas-phase molecule (in amu·Å^2). 1 element for linear molecules, 3 elements for
        non-linear molecules. Can be obtained from ase.Atoms.get_moments_of_inertia()
    sym_number: int
        Symmetry number of the molecule
    degeneracy: int
        Degeneracy of the ground state, for the calculation of the electronic partition function. Default value: 1

    Returns
    -------
    pe_fwd: float
        Pre-exponential factor in the forward direction (in s^-1·bar^-1)
    pe_rev: float
        Pre-exponential factor in the reverse direction (in s^-1)

    """
    area_site = area_site * 1.0e-20  # Å^2 to m^2
    m = molec_mass * 1.0e-3 / N_A  # g/mol to kg/molec
    q_vib_gas = get_q_vib(temperature=temperature, vib_energies=vib_energies_is)
    q_rot_gas = get_q_rot(temperature=temperature, inertia_moments=inertia_moments, sym_number=sym_number)
    q_trans_2d_gas = area_site * 2 * pi * m * k * temperature / h ** 2
    q_el_gas = degeneracy
    q_vib_ads = get_q_vib(temperature=temperature, vib_energies=vib_energies_fs)
    if not vib_energies_ts:  # non-activated
        pe_fwd = area_site / sqrt(2 * pi * m * k * temperature) * 1e5  # Pa-1 to bar-1
        pe_rev = (q_el_gas * q_vib_gas * q_rot_gas * q_trans_2d_gas / q_vib_ads) * (k * temperature / h)
    else:  # activated
        q_vib_ts = get_q_vib(temperature=temperature, vib_energies=vib_energies_ts)
        pe_fwd = (q_vib_ts / (q_el_gas * q_vib_gas * q_rot_gas * q_trans_2d_gas)) * (area_site / sqrt(2 * pi * m * k * temperature))
        pe_fwd = pe_fwd * 1e5  # Pa-1 to bar-1
        pe_rev = (q_vib_ts / q_vib_ads) * (k * temperature / h)
    return pe_fwd, pe_rev


def calc_surf_proc(temperature, vib_energies_is, vib_energies_ts, vib_energies_fs):
    """Calculates the forward and reverse pre-exponential factors for a reversible surface process.

    Parameters
    ----------
    temperature: float
        Temperature (in K)
    vib_energies_is: list
        Vibrational energies for the initial state (in meV)
    vib_energies_ts: list
        Vibrational energies for the transition state (in meV)
    vib_energies_fs: list
        Vibrational energies for the final state (in meV)

    Returns
    -------
    pe_fwd: float
        Pre-exponential factor in the forward direction (in s^-1)
    pe_rev: float
        Pre-exponential factor in the reverse direction (in s^-1)

    """
    q_vib_initial = get_q_vib(temperature=temperature, vib_energies=vib_energies_is)
    q_vib_ts = get_q_vib(temperature=temperature, vib_energies=vib_energies_ts)
    q_vib_final = get_q_vib(temperature=temperature, vib_energies=vib_energies_fs)
    pe_fwd = (q_vib_ts / q_vib_initial) * (k * temperature / h)
    pe_rev = (q_vib_ts / q_vib_final) * (k * temperature / h)
    return pe_fwd, pe_rev



import sys
import numpy as np
from zacrostools.analysis_functions import get_data_general, get_data_specnum


class KMCOutput:
    """A class that represents a KMC output

    Parameters
    ----------
    path: str
        Path of the directory containing the output files.
    ignore: float, optional
        Ignore first % of simulated time, i.e., equilibration (in %).
        Default value: 0.0
    coverage_per_site: bool, optional
        If True, calculate the coverage per site type.
        Default value: False
    ads_sites: dict, optional
        A dictionary where the surface species are stored as keys and their values are the site types where they adsorb,
        e.g. {'CO': 'top', 'O': 'hollow'}
        Default value: None


    Attributes
    ----------
    n_gas_species: int
        Number of gas species.
    gas_species_names: list of str
        Gas species names.
    n_surf_species: int
        Number of surface species.
    surf_species_names: list of str
        Surface species names.
    n_sites: int
        Total number of lattice sites.
    area: float
        Lattice surface area (in Å^2)
    site_types: dict
        Site type names and total number of sites of that type
    time: np.Array
        Simulated time (in s).
    final_time: float
        Final simulated time (in s).
    energy: np.Array
        Lattice energy (in eV·Å^-2).
    av_energy: float
        Average lattice energies (in eV·Å^-2).
    final_energy: float
        Final lattice energy (in eV·Å^-2).
    production: dict
        Gas species produced. Example: KMCOutput.production['CO']
    total_production: dict
        Total number of gas species produced. Example: KMCOutput.total_production['CO']
    tof: dict
        TOF of gas species (in molec·s^-1·Å^-2). Example: KMCOutput.tof['CO2']
    coverage: dict
        Coverage of surface species (in %). Example: KMCOutput.coverage['CO']
    av_coverage: dict
        Average coverage of surface species (in %). Example: KMCOutput.av_coverage['CO']
    total_coverage: np.Array
        Total coverage of surface species (in %).
    av_total_coverage: float
        Average total coverage of surface species (in %).
    dominant_ads: str
        Most dominant surface species, to plot the kinetic phase diagrams.
    coverage_per_site_type: dict
        Coverage of surface species per site type (in %).
    av_coverage_per_site_type: dict
        Average coverage of surface species per site type (in %).
    total_coverage_per_site_type: dict
        Total coverage of surface species per site type (in %). Example: KMCOutput.total_coverage_per_site_type['top']
    av_total_coverage_per_site_type: dict
        Average total coverage of surface species per site type (in %).
    dominant_ads_per_site_type: dict
        Most dominant surface species per site type, to plot the kinetic phase diagrams.
    """

    def __init__(self, path, ignore=0.0, coverage_per_site=False, ads_sites=None):
        self.path = path

        # Get data from general_output.txt file
        data_general = get_data_general(path)
        self.n_gas_species = data_general['n_gas_species']
        self.gas_species_names = data_general['gas_species_names']
        self.n_surf_species = data_general['n_surf_species']
        self.surf_species_names = data_general['surf_species_names']
        self.n_sites = data_general['n_sites']
        self.area = data_general['area']
        self.site_types = data_general['site_types']

        # Get data from specnum_output.txt
        data_specnum, header = get_data_specnum(path, ignore)
        self.time = data_specnum[:, 2]
        self.final_time = data_specnum[-1, 2]
        self.energy = data_specnum[:, 4] / self.area
        self.av_energy = np.average(data_specnum[:, 4]) / self.area
        self.final_energy = data_specnum[-1, 4] / self.area

        # Production (molec) and TOF (molec·s^-1·Å^-2)
        self.production = {}
        self.total_production = {}  # useful when calculating selectivity (i.e., set min_total_production)
        self.tof = {}
        for i in range(5 + self.n_surf_species, len(header)):
            ads = header[i]
            self.production[ads] = data_specnum[:, i]
            self.total_production[ads] = data_specnum[-1, i] - data_specnum[0, i]
            if data_specnum[-1, i] != 0:
                self.tof[header[i]] = np.polyfit(data_specnum[:, 2], data_specnum[:, i], 1)[0] / self.area
            else:
                self.tof[header[i]] = 0.00

        # Coverage (%)
        self.coverage = {}
        self.av_coverage = {}
        for i in range(5, 5 + self.n_surf_species):
            ads = header[i].replace('*', '')
            self.coverage[ads] = data_specnum[:, i] / self.n_sites * 100
            self.av_coverage[ads] = np.average(data_specnum[:, i]) / self.n_sites * 100
        self.total_coverage = sum(self.coverage.values())
        self.av_total_coverage = sum(self.av_coverage.values())
        self.dominant_ads = max(self.av_coverage, key=self.av_coverage.get)

        # Coverage per site type (%)
        if len(self.site_types) == 1:
            print("ERROR: coverage_per_site not available when there is only one site type")
            sys.exit(f"path: {self.path}")
        if coverage_per_site:
            self.coverage_per_site_type = {}
            self.av_coverage_per_site_type = {}
            for site_type in self.site_types:
                self.coverage_per_site_type[site_type] = {}
                self.av_coverage_per_site_type[site_type] = {}
            for i in range(5, 5 + self.n_surf_species):
                ads = header[i].replace('*', '')
                site_type = ads_sites[ads]
                self.coverage_per_site_type[site_type][ads] = data_specnum[:, i] / self.site_types[ads_sites[ads]] * 100
                self.av_coverage_per_site_type[site_type][ads] = np.average(data_specnum[:, i]) / self.site_types[
                    ads_sites[ads]] * 100
            self.total_coverage_per_site_type = {}
            self.av_total_coverage_per_site_type = {}
            self.dominant_ads_per_site_type = {}
            for site_type in self.site_types:
                self.total_coverage_per_site_type[site_type] = sum(self.coverage_per_site_type[site_type].values())
                self.av_total_coverage_per_site_type[site_type] = sum(
                    self.av_coverage_per_site_type[site_type].values())
                self.dominant_ads_per_site_type[site_type] = max(self.av_coverage_per_site_type[site_type],
                                                                 key=self.av_coverage_per_site_type[site_type].get)

    def get_selectivity(self, main_product, side_products):
        selectivity = float('NaN')
        tof_side_products = 0.0
        for side_product in side_products:
            tof_side_products += self.tof[side_product]
        if self.tof[main_product] + tof_side_products != 0:
            selectivity = self.tof[main_product] / (self.tof[main_product] + tof_side_products) * 100
        return selectivity



import os
from random import randint
from zacrostools.input_functions import write_header
from zacrostools.mechanism_input import ReactionModel
from zacrostools.energetics_input import EnergeticModel


class KMCModel:
    """A class that represents a KMC model.

    Parameters
    ----------
    gas_data: pd.DataFrame
        A Pandas DataFrame containing informaton about the gas molecules.
    mechanism_data: pd.DataFrame
        A Pandas DataFrame containing informaton about the reaction model.
    energetics_data: pd.DataFrame
        A Pandas DataFrame containing informaton about the energetic model.
    lattice_model: zacrostools.lattice_input.LatticeModel
        A lattice model
    """

    def __init__(self, gas_data, mechanism_data, energetics_data, lattice_model):
        self.path = None
        self.gas_data = gas_data
        self.reaction_model = ReactionModel(mechanism_data=mechanism_data)
        self.energetic_model = EnergeticModel(energetics_data=energetics_data)
        self.lattice_model = lattice_model

    def create_job_dir(self, path, temperature, pressure, report='on event 10000', stop=None,
                       manual_scaling=None, auto_scaling=None, auto_scaling_tags=None):
        """

        Parameters
        ----------
        path: str
            The path for the job directory where input files will be written
        temperature: float
            Reaction temperature (in K)
        pressure: dict
            Partial pressures of all gas species (in bar), e.g. {'CO': 1.0, 'O2': 0.001}
        report: str, optional
            Reporting scheme in Zacros format. Default value: 'on event 100000'
        stop: dict, optional
            Stopping criteria in Zacros format. Must contain the following keys: 'max_steps', 'max_time' and 'wall_time'
            Default value: {'max_steps': 'infinity', 'max_time': 'infinity', 'wall_time': 86400}
        manual_scaling: list of str, optional
            List of processes to downscale manually, e.g. ['CO_diffusion', 'O_diffusion']
            Default value: []
        auto_scaling: list of str, optional
            List of processes that will be marked as 'stiffness_scalable' in mechanism_input.dat.
            Default value: []
        auto_scaling_tags: dict, optional
            Dictionary of keywords controling the dynamic scaling algorithm and the corresponding values,
            e.g. {'check_every': 500, 'min_separation': 400.0, 'max_separation': 600.0}
            Default value: {}

        """
        if stop is None:
            stop = {'max_steps': 'infinity', 'max_time': 'infinity', 'wall_time': 86400}
        if manual_scaling is None:
            manual_scaling = {}
        if auto_scaling is None:
            auto_scaling = []
        if auto_scaling_tags is None:
            auto_scaling_tags = {}
        if len(auto_scaling) == 0 and len(auto_scaling_tags) > 0:
            print('WARNING: auto_scaling_tags provided but dynamic scaling is not enabled')
        self.path = path
        if not os.path.exists(self.path):
            os.mkdir(self.path)
            self.write_simulation_input(temperature=temperature, pressure=pressure, report=report, stop=stop,
                                        auto_scaling=auto_scaling, auto_scaling_tags=auto_scaling_tags)
            self.reaction_model.write_mechanism_input(path=self.path, temperature=temperature, gas_data=self.gas_data,
                                                      manual_scaling=manual_scaling, auto_scaling=auto_scaling)
            self.energetic_model.write_energetics_input(path=self.path)
            self.lattice_model.write_lattice_input(path=self.path)
        else:
            print(f'{self.path} already exists (nothing done)')

    def write_simulation_input(self, temperature, pressure, report, stop, auto_scaling, auto_scaling_tags):
        """Writes the simulation_input.dat file"""
        gas_specs_names = [x for x in self.gas_data.index]
        surf_specs_names = [x.replace('_point', '') for x in self.energetic_model.df.index if '_point' in x]
        surf_specs_names = [x + '*' * int(self.energetic_model.df.loc[f'{x}_point', 'sites']) for x in surf_specs_names]
        surf_specs_dent = [x.count('*') for x in surf_specs_names]
        write_header(f"{self.path}/simulation_input.dat")
        with open(f"{self.path}/simulation_input.dat", 'a') as infile:
            infile.write('random_seed\t'.expandtabs(26) + str(randint(100000, 999999)) + '\n')
            infile.write('temperature\t'.expandtabs(26) + str(float(temperature)) + '\n')
            p_tot = sum(pressure.values())
            infile.write('pressure\t'.expandtabs(26) + str(p_tot) + '\n')
            infile.write('n_gas_species\t'.expandtabs(26) + str(len(gas_specs_names)) + '\n')
            infile.write('gas_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in gas_specs_names) + '\n')
            tags_dict = ['gas_energy', 'gas_molec_weight']
            tags_zacros = ['gas_energies', 'gas_molec_weights']
            for tag1, tag2 in zip(tags_dict, tags_zacros):
                tag_list = [self.gas_data.loc[x, tag1] for x in gas_specs_names]
                infile.write(f'{tag2}\t'.expandtabs(26) + " ".join(str(x) for x in tag_list) + '\n')
            gas_molar_frac_list = [pressure[x] / p_tot for x in gas_specs_names]
            infile.write(f'gas_molar_fracs\t'.expandtabs(26) + " ".join(str(x) for x in gas_molar_frac_list) + '\n')
            infile.write('n_surf_species\t'.expandtabs(26) + str(len(surf_specs_names)) + '\n')
            infile.write('surf_specs_names\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs_names) + '\n')
            infile.write('surf_specs_dent\t'.expandtabs(26) + " ".join(str(x) for x in surf_specs_dent) + '\n')
            for tag in ['snapshots', 'process_statistics', 'species_numbers']:
                infile.write((tag + '\t').expandtabs(26) + report + '\n')
            for tag in ['max_steps', 'max_time', 'wall_time']:
                infile.write((tag + '\t').expandtabs(26) + str(stop[tag]) + '\n')
            if len(auto_scaling) > 0:
                infile.write(f"enable_stiffness_scaling\n")
                for tag in auto_scaling_tags:
                    infile.write((tag + '\t').expandtabs(26) + str(auto_scaling_tags[tag]) + '\n')
            infile.write(f"finish\n")



import ast
import sys
import pandas as pd
from zacrostools.input_functions import write_header, calc_ads, calc_surf_proc


class ReactionModel:
    """A class that represents a KMC reaction model.

    Parameters:

    mechanism_data: Pandas DataFrame
        Information on the reaction model
        The reaction name is taken as the index of each row

        The following columns are required:
            - site_types (str): the types of each site in the pattern
            - initial (list): initial configuration in Zacros format, e.g. ['1 CO* 1','2 * 1']
            - final (list): final configuration in Zacros format, e.g. ['1 C* 1','2 O* 1']
            - activ_eng (float): activation energy (in eV)
            - vib_energies_is (list of floats): vibrational energies for the initial state (in meV)
            - vib_energies_ts (list of floats): vibrational energies for the transition state (in meV).
              For non-activated adsorption, define this as an empty list i.e. []
            - vib_energies_fs (list of floats): vibrational energies for the final state (in meV)
            - molecule (str): gas-phase molecule involved. Only required for adsorption steps. Default value: None
            - area_site (float): area of adsorption site (in Å^2). Only required for adsorption steps. Default value: None
        The following columns are optional:
            - neighboring (str): connectivity between sites involved, e.g. 1-2. Default value: None
            - prox_factor (float): proximity factor. Default value: 0.5
            - angles (str): Angle between sites in Zacros format, e.g. '1-2-3:180'. Default value: None
    """

    def __init__(self, mechanism_data=None):
        if isinstance(mechanism_data, pd.DataFrame):
            self.df = mechanism_data
        else:
            print("Error: parameter 'mechanism_data' in ReactionModel is not a Pandas DataFrame")

    @classmethod
    def from_dictionary(cls, path):
        """Not implemented yet"""
        pass

    def write_mechanism_input(self, path, temperature, gas_data, manual_scaling, auto_scaling):
        """Writes the mechanism_input.dat file"""
        write_header(f"{path}/mechanism_input.dat")
        with open(f"{path}/mechanism_input.dat", 'a') as infile:
            infile.write('mechanism\n\n')
            infile.write('############################################################################\n\n')
            for step in self.df.index:
                initial_state = ast.literal_eval(self.df.loc[step, 'initial'])
                final_state = ast.literal_eval(self.df.loc[step, 'final'])
                if len(initial_state) != len(final_state):
                    sys.exit(f"Error in {step}: len IS is {len(initial_state)} but len FS is {len(final_state)}")
                infile.write(f"reversible_step {step}\n\n")
                if not pd.isna(self.df.loc[step, 'molecule']):
                    infile.write(f"  gas_reacs_prods {self.df.loc[step, 'molecule']} -1\n")
                infile.write(f"  sites {len(initial_state)}\n")
                if not pd.isna(self.df.loc[step, 'neighboring']):
                    infile.write(f"  neighboring {self.df.loc[step, 'neighboring']}\n")
                infile.write(f"  initial\n")
                for element in initial_state:
                    infile.write(f"    {element}\n")
                infile.write(f"  final\n")
                for element in final_state:
                    infile.write(f"    {element}\n")
                infile.write(f"  site_types {self.df.loc[step, 'site_types']}\n")
                pre_expon, pe_ratio = self.get_pre_expon(step=step, temperature=temperature, gas_data=gas_data,
                                                         manual_scaling=manual_scaling)
                if step in manual_scaling:
                    infile.write(f"  pre_expon {pre_expon:.3e}   # scaled {manual_scaling[step]:.3e}\n")
                else:
                    infile.write(f"  pre_expon {pre_expon:.3e}\n")
                infile.write(f"  pe_ratio {pe_ratio:.3e}\n")
                infile.write(f"  activ_eng {self.df.loc[step, 'activ_eng']:.2f}\n")
                for keyword in ['prox_factor', 'angles']:  # optional keywords
                    if keyword in self.df.columns:
                        if not pd.isna(self.df.loc[step, keyword]):
                            infile.write(f"  {keyword} {self.df.loc[step, keyword]}\n")
                if step in auto_scaling:
                    infile.write(f"  stiffness_scalable \n")
                infile.write(f"\nend_reversible_step\n\n")
                infile.write('############################################################################\n\n')
            infile.write(f"end_mechanism\n")

    def get_pre_expon(self, step, temperature, gas_data, manual_scaling):
        """Calculates the forward pre-exponential and the pre-exponential ratio, required for the mechanism_input.dat
        file """
        vib_energies_is = ast.literal_eval(self.df.loc[step, 'vib_energies_is'])
        vib_energies_ts = ast.literal_eval(self.df.loc[step, 'vib_energies_ts'])
        vib_energies_fs = ast.literal_eval(self.df.loc[step, 'vib_energies_fs'])
        if not pd.isna(self.df.loc[step, 'molecule']):  # adsorption
            molecule = self.df.loc[step, 'molecule']
            molec_mass = gas_data.loc[molecule, 'gas_molec_weight']
            inertia_moments = ast.literal_eval(gas_data.loc[molecule, 'inertia_moments'])
            if 'degeneracy' in gas_data:  # test that it works
                if not pd.isna(gas_data.loc[molecule, 'degeneracy']):
                    degeneracy = int(gas_data.loc[molecule, 'degeneracy'])
                else:
                    degeneracy = 1
            else:
                degeneracy = 1
            pe_fwd, pe_rev = calc_ads(area_site=self.df.loc[step, 'area_site'],
                                      molec_mass=molec_mass,
                                      temperature=temperature,
                                      vib_energies_is=vib_energies_is,
                                      vib_energies_ts=vib_energies_ts,
                                      vib_energies_fs=vib_energies_fs,
                                      inertia_moments=inertia_moments,
                                      sym_number=int(gas_data.loc[molecule, 'sym_number']),
                                      degeneracy=degeneracy)
        else:  # surface process
            pe_fwd, pe_rev = calc_surf_proc(temperature=temperature,
                                            vib_energies_is=vib_energies_is,
                                            vib_energies_ts=vib_energies_ts,
                                            vib_energies_fs=vib_energies_fs)

        if step in manual_scaling:
            pe_fwd = pe_fwd * manual_scaling[step]
            pe_rev = pe_rev * manual_scaling[step]
        pe_ratio = pe_fwd / pe_rev
        return pe_fwd, pe_ratio



import numpy as np


def find_nearest(array, value):
    """Finds the element of an array whose value is closest to a given value, and returns its index.

    Parameters
    ----------
    array: np.Array
    value: float

    Returns
    -------
    array[idx]: int

    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def get_data_general(path):
    dmatch = ['Number of gas species',
              'Gas species names',
              'Number of surface species',
              'Surface species names',
              'Total number of lattice sites',
              'Lattice surface area',
              'Site type names and total number of sites of that type']
    data = {}
    with open(f"{path}/general_output.txt", 'r') as file_object:
        line = file_object.readline()
        while len(dmatch) != 0:
            if 'Number of gas species' in line:
                data['n_gas_species'] = int(line.split()[-1])
                dmatch.remove('Number of gas species')
            if 'Gas species names' in line:
                data['gas_species_names'] = line.split(':')[-1].split()
                dmatch.remove('Gas species names')
            if 'Number of surface species' in line:
                data['n_surf_species'] = int(line.split()[-1])
                dmatch.remove('Number of surface species')
            if 'Surface species names' in line:
                data['surf_species_names'] = [ads[0:-1] for ads in line.split(':')[-1].split()]
                dmatch.remove('Surface species names')
            if 'Total number of lattice sites' in line:
                data['n_sites'] = int(line.split()[-1])
                dmatch.remove('Total number of lattice sites')
            if 'Lattice surface area' in line:
                data['area'] = float(line.split()[-1])
                dmatch.remove('Lattice surface area')
            if 'Site type names and total number of sites of that type' in line:
                line = file_object.readline()
                site_types = {}
                while line.strip():
                    num_sites_of_given_type = int(line.strip().split(' ')[1].replace('(', '').replace(')', ''))
                    site_types[line.strip().split(' ')[0]] = num_sites_of_given_type
                    line = file_object.readline()
                data['site_types'] = site_types
                dmatch.remove('Site type names and total number of sites of that type')
            line = file_object.readline()
        return data


def get_data_specnum(path, ignore=0.0):
    with open(f"{path}/specnum_output.txt", "r") as infile:
        header = infile.readline().split()
    full_data = np.loadtxt(f"{path}/specnum_output.txt", skiprows=1)
    index = np.where(full_data[:, 2] == find_nearest(full_data[:, 2], full_data[-1, 2] * ignore / 100))[0][0]
    data = np.delete(full_data, slice(0, index), 0)
    return data, header


def get_data_simulation(path):
    dmatch = ['pressure',
              'gas_specs_names',
              'gas_molar_fracs']
    data = {}
    with open(f"{path}/simulation_input.dat", 'r') as file_object:
        line = file_object.readline()
        while len(dmatch) != 0:
            if 'pressure' in line:
                data['pressure'] = float(line.split()[-1])
                dmatch.remove('pressure')
            if 'gas_specs_names' in line:
                data['gas_specs_names'] = line.split()[1:]
                dmatch.remove('gas_specs_names')
            if 'gas_molar_fracs' in line:
                data['gas_molar_fracs'] = [float(x) for x in line.split()[1:]]
                dmatch.remove('gas_molar_fracs')
            line = file_object.readline()
        return data
